-- (c) 2007 Nymbia.  see LGPLv2.1.txt for full details.
--this tool is run in the lua command line.  http://lua.org
--socket is required for internet data.
--get socket here: http://luaforge.net/projects/luasocket/

local SOURCE = SOURCE or "data.lua"
local DEBUG = DEBUG or 1

local function dprint(dlevel, ...)
	if dlevel and DEBUG >= dlevel then
		print(...)
	end
end
local sets

local url = require("socket.url")
local httptime, httpcount = 0, 0
local getpage
do
	local status, curl = pcall(require, "curl")
	if status then
		local temp
		local write = function (s, len)
			temp[#temp + 1] = s
			return len
		end
		local c = curl.easy_init()
		function getpage(url)
			dprint(2, "curl", url)
			temp = {}
			c:setopt(curl.OPT_URL, url)
			c:setopt(curl.OPT_WRITEFUNCTION, write)
			local stime = os.time()
			local err, info = c:perform()
			httptime = httptime + (os.time() - stime)
			httpcount = httpcount + 1
			if err ~= 0 then
				dprint(1, "curl error", url, info)
			else
				temp = table.concat(temp)
				if temp:len() > 0 then
					return temp
				end
			end
		end
	else
		local http = require("socket.http")

		function getpage(url)
			dprint(2, "socket.http", url)
			local stime = os.time()
			local r = http.request(url)
			httptime = httptime + (os.time() - stime)
			httpcount = httpcount + 1
			return r
		end
	end
end

local function read_data_file()
	local subset = arg[1] or ''
	local f = assert(io.open(SOURCE, "r"))
	local file = f:read("*all")
	f:close()

	local sets = {}
	local setcount = 0
	print('\t%[%"('..subset..'[^"]+)%"%][^=]-= "([^"]-)"')
	for set, data in file:gmatch('\t%[%"('..subset..'[^"]+)%"%][^=]-= "([^"]-)"') do
		sets[set] = data
		setcount = setcount + 1
	end
	return file, sets, setcount
end

local handlers = {}

local function basic_itemid_handler(itemstring)
	return itemstring:match("{id:(%d+)")
end

local function basic_listview_handler(url, handler)
	if not handler then handler = basic_itemid_handler end
	local page = getpage(url)
	local itemlist = page:match("new Listview(%b())"):match("data%: (%b[])")
	local newset
	for itemstring in itemlist:gmatch("%b{}") do
		local s = handler(itemstring)
		if s then
			if newset then
				newset = newset..","..s
			else
				newset = s
			end
		end
	end
	return newset
end

local function basic_listview_get_first_id(url)
	local page = getpage(url)
	if not page then return end
	page = page:match("new Listview(%b())")
	if not page then return end
	page = page:match("data: (%b[])")
	if not page then return end
	page = page:match("{id:(%d+)")
	if not page then return end
	return tonumber(page)
end

local Tradeskill_Gather_filters = {
	Disenchant = 68,
	Fishing = 69,
	Herbalism = 70,
	Mining = 73,
	Pickpocketing = 75,
	Skinning = 76,
	Prospecting = 88,
}

local Tradeskill_Tool_urls = {
	Alchemy = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=12;crv=0", -- Tool - Philosopher's Stone
	},
	Blacksmithing = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=13;crv=0", -- Tool - Blacksmith Hammer
	},
	Enchanting = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=62;crv=0", -- Tool - Runed Adamantite Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=10;crv=0", -- Tool - Runed Arcanite Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=6;crv=0",  -- Tool - Runed Copper Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=63;crv=0", -- Tool - Runed Eternium Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=41;crv=0", -- Tool - Runed Fel Iron Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=8;crv=0",  -- Tool - Runed Golden Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=7;crv=0",  -- Tool - Runed Silver Rod
		"http://www.wowhead.com/?items&filter=cr=91;crs=9;crv=0",  -- Tool - Runed Truesilver Rod
	},
	Engineering = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=14;crv=0", -- Tool - Arclight Spanner
		"http://www.wowhead.com/?items&filter=cr=91;crs=13;crv=0", -- Tool - Blacksmith Hammer
		"http://www.wowhead.com/?items&filter=cr=91;crs=15;crv=0", -- Tool - Gyromatic Micro-Adjustor
	},
	Mining = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=11;crv=0", -- Tool - Mining Pick
	},
	Skinning = {
		"http://www.wowhead.com/?items&filter=cr=91;crs=1;crv=0",  -- Tool - Skinning Knife
	},
}

local Reagent_Ammo_Filters = {
	Arrow = "6.2",
	Bullet = "6.3",
	Thrown = "2.16",
}

local Containers_ItemsInType_items = {
	["Soul Shard"] = 21342,
	Herb = 22251,
	Enchanting = 21858,
	Engineering = 30745,
	Gem = 30747,
	Mining = 29540,
}

local Bag_filters = {
	Basic = "1.0",
	["Soul Shard"] = "1.1",
	Herb = "1.2",
	Enchanting = "1.3",
	Engineering = "1.4",
	Jewelcrafting = "1.5",
	Mining = "1.6",
	Ammo = "11.3",
	Quiver = "11.2",
}

local Tradeskill_Recipe_filters = {
	Leatherworking = 1,
	Tailoring = 2,
	Engineering = 3,
	Blacksmithing = 4,
	Cooking = 5,
	Alchemy = 6,
	["First Aid"] = 7,
	Enchanting = 8,
	Fishing = 9,
	Jewelcrafting = 10,
}

local Tradeskill_Recipe_filters = {
	Quest = "18;crs=1;crv=0",
	Drop = "72;crs=1;crv=0",
	Crafted = "86;crs=11;crv=0",
	Vendor = "92;crs=1;crv=0",
	Other = "18:72:86:92;crs=5:2:12:2;crv=0:0:0:0",
}

local Tradeskill_Gather_GemsInNodes_nodes = {
	["Copper Vein"] = 1731,
	["Tin Vein"] = 1732,
	["Silver Vein"] = 1733,
	["Iron Deposit"] = 1735,
	["Gold Vein"] = 1734,
	["Mithril Deposit"] = 2040,
	["Truesilver Deposit"] = 2047,
	["Small Thorium Vein"] = 324,
	["Hakkari Thorium Vein"] = 180215,
	["Rich Thorium Vein"] = 175404,
	["Dark Iron Deposit"] = 165658,
	["Fel Iron Deposit"] = 181555,
	["Adamantite Deposit"] = 181556,
	["Rich Adamantite Deposit"] = 181569,
	["Khorium Vein"] = 181557,
}

local Tradeskill_Profession_filters = {
	Alchemy = "11.171",
	["Blacksmithing.Basic"] = "11.164&filter=cr=5;crs=2;crv=0",
	["Blacksmithing.Armorsmith"] = "11.164.9788",
	["Blacksmithing.Weaponsmith.Axesmith"] = "11.164.17041",
	["Blacksmithing.Weaponsmith.Hammersmith"] = "11.164.17040",
	["Blacksmithing.Weaponsmith.Swordsmith"] = "11.164.17039",
	["Blacksmithing.Weaponsmith.Basic"] = "11.164.9787",
	Cooking = "9.185",
	Enchanting = "11.333",
	["Engineering.Basic"] = "11.202&filter=cr=5;crs=2;crv=0",
	["Engineering.Gnomish"] = "11.202.20219",
	["Engineering.Goblin"] = "11.202.20222",
	["First Aid"] = "9.129",
	Jewelcrafting = "11.755",
	["Leatherworking.Basic"] = "11.165&filter=cr=5;crs=2;crv=0",
	["Leatherworking.Dragonscale"] = "11.165.10656",
	["Leatherworking.Elemental"] = "11.165.10658",
	["Leatherworking.Tribal"] = "11.165.10660",
	Smelting = "11.186",
	["Tailoring.Basic"] = "11.197&filter=cr=5;crs=2;crv=0",
	["Tailoring.Mooncloth"] = "11.197.26798",
	["Tailoring.Shadoweave"] = "11.197.26801",
	["Tailoring.Spellfire"] = "11.197.26797",
	Poisons = "7.4.40",
}

local Gear_Socketed_urls = {
	Back	= {
		"http://www.wowhead.com/?items&filter=sl=16;cr=80;crs=5;crv=0",
	},
	Chest	= {
		"http://www.wowhead.com/?items&filter=sl=5;cr=72:80;crs=1:5;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=5;cr=93:80;crs=1:5;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=5;cr=93:80:72;crs=2:5:2;crv=0:0:0",
	},
	Feet	= {
		"http://www.wowhead.com/?items&filter=sl=8;cr=80;crs=5;crv=0",
	},
	Finger	= {
		"http://www.wowhead.com/?items&filter=sl=11;cr=80;crs=5;crv=0",
	},
	Hands	= {
		"http://www.wowhead.com/?items&filter=sl=10;cr=80;crs=5;crv=0",
	},
	Head	= {
		"http://www.wowhead.com/?items&filter=sl=1;cr=80:72;crs=5:1;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=1;cr=80:93;crs=5:1;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=1;cr=80:93:72;crs=5:2:2;crv=0:0:0",
	},
	Legs	= {
		"http://www.wowhead.com/?items&filter=sl=7;cr=80;crs=5;crv=0",
	},
	["Main Hand"]	= {
		"http://www.wowhead.com/?items&filter=sl=21;cr=80;crs=5;crv=0",
	},
	Neck	= {
		"http://www.wowhead.com/?items&filter=sl=2;cr=80;crs=5;crv=0",
	},
	["One Hand"]	= {
		"http://www.wowhead.com/?items&filter=sl=13;cr=80;crs=5;crv=0",
	},
	Ranged	= {
		"http://www.wowhead.com/?items&filter=sl=15;cr=80;crs=5;crv=0",
	},
	Shield	= {
		"http://www.wowhead.com/?items&filter=sl=14;cr=80;crs=5;crv=0",
	},
	Shoulder	= {
		"http://www.wowhead.com/?items&filter=sl=3;cr=80:72;crs=5:1;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=3;cr=80:93;crs=5:1;crv=0:0",
		"http://www.wowhead.com/?items&filter=sl=3;cr=80:93:72;crs=5:2:2;crv=0:0:0",
	},
	["Two Hand"]	= {
		"http://www.wowhead.com/?items&filter=sl=17;cr=80;crs=5;crv=0",
	},
	Waist	= {
		"http://www.wowhead.com/?items&filter=sl=6;cr=80;crs=5;crv=0",
	},
	Wrist	= {
		"http://www.wowhead.com/?items&filter=sl=9;cr=80;crs=5;crv=0",
	},
}

local Tradeskill_Gem_Color_filters = {
	Red = 0,
	Blue = 1,
	Yellow = 2,
	Purple = 3,
	Green = 4,
	Orange = 5,
	Meta = 6,
	-- Simple = 7,
	Prismatic = 8
}

local Consumable_Bandage_filters = {
	Basic = "cr=86;crs=6;crv=0",
	["Alterac Valley"] = "na=bandage;cr=92:104;crs=1:0;crv=0:Alterac",
	["Warsong Gulch"] = "na=bandage;cr=92:107;crs=1:0;crv=0:Warsong",
	["Arathi Basin"] = "na=bandage;cr=92:107;crs=1:0;crv=0:Arathi",
}

local Consumable_Buff_Type_filters = {
	["Battle"] = "cr=107;crs=0;crv=battle+elixir",
	["Both"] = "cr=107;crs=0;crv=battle+guardian+elixir",
	["Guardian"] = "cr=107;crs=0;crv=guardian+elixir",
}	

handlers["^InstanceLoot%."] = function (set, data) -- fixed
	local newset
	local boss = set:match("%.([^%.]+)$")
	if boss == " Smite" then
		boss = "Mr. Smite"
	end
	local id, type = 
		basic_listview_get_first_id("http://www.wowhead.com/?npcs&filter=na="..url.escape(boss)), "npc"
	if not id then
		id, type = basic_listview_get_first_id("http://www.wowhead.com/?objects&filter=na="..url.escape(boss)), "object"
	end
	if id then
		page = getpage("http://www.wowhead.com/?"..type.."="..id)
		local count = 0
		local drops = {}
		local normaldropslist, heroicdropslist, dropslist
		for list in page:gmatch("new Listview(%b())") do
			local id = list:match("id: '([^']+)'")
			local data = list:match("data%: (%b[])")
			local count = tonumber(list:match("note:[^,]+, (%d+)"))
			dprint(4, id, count)
			if
				id == "contains" or
				id == "normal-drops" or
				id == "drops" or
				id == "normal-contents"
			then
				drops.normal = { data, count }
			elseif id == "heroic-drops" or id == "heroic-contents"	then
				drops.heroic = { data, count }
			end
		end
		local totaldrops
		local handler = function (itemstring)
			local itemid = itemstring:match("{id:(%d+)")
			local dropcount = tonumber(itemstring:match("count:(%d+)"))
			local droprate = math.floor(dropcount / totaldrops * 1000)
			if droprate > 50 then
				return itemid..":"..droprate
			end
		end
		if drops.normal then
			local itemlist
			itemlist, totaldrops = unpack(drops.normal)
			for itemstring in itemlist:gmatch("%b{}") do
				local v = handler(itemstring)
				if v then
					if newset then
						newset = newset..","..v
					else
						newset = v
					end
					count = count + 1
				end
			end
		else
			dprint(2, "no drops for "..boss)
		end
		if drops.heroic then
			local normalsub = set:match("^InstanceLoot%.(.+)$")
			local heroicname = "InstanceLootHeroic."..normalsub
			local heroicstr, itemlist
			itemlist, totaldrops = unpack(drops.heroic)
			for itemstring in itemlist:gmatch("%b{}") do
				local v = handler(itemstring)
				if v then
					if heroicstr then
						heroicstr = heroicstr..","..v
					else
						heroicstr = v
					end
				end
			end
			sets[heroicname] = heroicstr
		end
		dprint(2, "InstanceLoot: "..boss.." has "..count)
	end
	return newset
end

handlers["^GearSet"] = function (set, data) -- fixed
	local newset
	local setname = set:match("%.([^%.]+)$")
	local id = basic_listview_get_first_id("http://www.wowhead.com/?itemsets&filter=na="..url.escape(setname))
	if id then
		local count = 0
		page = getpage("http://www.wowhead.com/?itemset="..id)
		for itemstring in page:gmatch("ge%('iconlist%-icon%d+'%)[^\n]+") do
			local itemid = itemstring:match("%.createIcon%((%d+)")
			if itemid then
				if newset then
					newset = newset..","..itemid
				else
					newset = itemid
				end
				count = count + 1
			else
				error("no itemid")
			end
		end
		dprint(2, "GearSet: "..setname.." has "..count)
	end
	return newset
end

handlers["^Reagent%.Ammo"] = function (set, data) -- fixed
	local newset
	local setname = set:match("%.([^%.]+)$")
	local count = 0
	local filter = Reagent_Ammo_Filters[setname]
	if not filter then return end
	newset = basic_listview_handler("http://www.wowhead.com/?items="..filter, function (itemstring)
		count = count + 1
		local itemid = itemstring:match("{id:(%d+)")
		local dps = itemstring:match("dps:(%d+%.%d+)")
		dps = math.floor(tonumber(dps)*10)
		return itemid..":"..dps
	end)
	dprint(2, "Reagent.Ammo."..setname..":"..count)
	return newset
end

handlers["^Bag%."] = function (set, data) -- fixed
	local setname = set:match("%.([^%.]+)$")
	local count = 0
	local searchstring = Bag_filters[setname]
	if not searchstring then return end
	return basic_listview_handler("http://www.wowhead.com/?items="..searchstring, function (itemstring)
		count = count + 1
		local itemid = itemstring:match("{id:(%d+)")
		local nslots = itemstring:match("nslots:(%d+)")
		return itemid..":"..nslots
	end)
end

handlers["^Consumable%.Bandage"] = function (set, data) -- fixed
	local newset
	local setname = set:match("%.([^%.]+)$")
	local filter = Consumable_Bandage_filters[setname]
	if not filter then return end
	local page = getpage("http://www.wowhead.com/?items&filter="..filter)
	for itemid, content in page:gmatch("_%[(%d+)%]=(%b[])") do
		local heal = content:match("Heals (%d+) damage")
		if heal then
			if newset then
				newset = newset..","..itemid..":"..heal
			else
				newset = itemid..":"..heal
			end
		end
	end
	return newset
end

handlers["^Consumable%.Buff Type"] = function (set, data) -- fixed
	local newset
	local setname = set:match("%.([^%.]+)$")
	local both

	local filter = Consumable_Buff_Type_filters[setname]
	if not filter then return end

	local page = getpage("http://www.wowhead.com/?items&filter="..Consumable_Buff_Type_filters['Both'])

	for itemid, content in page:gmatch("_%[(%d+)%]=(%b[])") do
		if both then
			both = both..","..itemid
		else
			both = itemid
		end
	end
	
	if setname == 'Both' then return both end

	local page = getpage("http://www.wowhead.com/?items&filter="..filter)
	for itemid, content in page:gmatch("_%[(%d+)%]=(%b[])") do
		if not both:match( itemid ) then
			if newset then
				newset = newset..","..itemid
			else
				newset = itemid
			end
		end
	end
	return newset
end

handlers["^Tradeskill%.Gather%.GemsInNodes"] = function (set, data) -- fixed
	local nodetype = set:match("%.([^%.]+)$")
	local count = 0
	dprint(9, nodetype)
	local id = Tradeskill_Gather_GemsInNodes_nodes[nodetype]
	if not id then return end
	return basic_listview_handler("http://www.wowhead.com/?object="..id, function(itemstring)
		local itemid = itemstring:match("{id:(%d+)")
		local class = itemstring:match("classs?:(%d+)")
		if class == "3" then return itemid end
	end)
end


handlers["^Tradeskill%.Crafted"] = function (set, data) -- fixed
	local newset
	local profession = set:match("^Tradeskill%.Crafted%.(.+)$")
	dprint(9, "profession", profession)
	local filter = Tradeskill_Profession_filters[profession]
	if not filter then return end
	
	local fp_set, rp_set, lp_set = nil, "o", nil
	
	
	local reagenttable = {}
	basic_listview_handler("http://www.wowhead.com/?spells="..filter, function (itemstring)
		local spellid = itemstring:match("%{id:(%d+)")
		local skilllvl = itemstring:match("learnedat:(%d+)")
		local reagentstring = itemstring:match("reagents:(%b[])")
		if not reagentstring then return end
		skilllvl = math.min(375, tonumber(skilllvl) or 375)
		local itemid
		dprint(3, profession, itemid, skilllvl, reagentstring)
		if profession == "Enchanting" then
			itemid = -1 * spellid
			if itemid and skilllvl > 0 then
				if newset then
					newset = newset..","..itemid..":"..skilllvl
				else
					newset = itemid..":"..skilllvl
				end
				local newrecipemats = itemid..":"
				for reagentid, reagentnum in reagentstring:sub(2,-2):gmatch("%[(%d+),(%d+)%]") do
					if reagenttable[reagentid] then
						reagenttable[reagentid] = reagenttable[reagentid]..";"..itemid.."x"..reagentnum
					else
						reagenttable[reagentid] = itemid.."x"..reagentnum
					end
					newrecipemats = newrecipemats..reagentid.."x"..reagentnum..";"
				end
				newrecipemats = newrecipemats:sub(1,-2)
				if fp_set then
					fp_set = fp_set..","..newrecipemats
				else
					fp_set = "o,"..newrecipemats
				end
				if lp_set then
					lp_set = lp_set..",-"..spellid..":"..skilllvl
				else
					lp_set = "-"..spellid..":"..skilllvl
				end
			end
		else
			itemid = itemstring:match("creates:%[(%d+),(%d+),(%d+)%]") -- count ?
			if itemid and skilllvl > 0 then
				if newset then
					newset = newset..","..itemid..":"..skilllvl
				else
					newset = itemid..":"..skilllvl
				end
				local newrecipemats = itemid..":"
				for reagentid,reagentnum in reagentstring:sub(2,-2):gmatch("%[(%d+),(%d+)%]") do
					if reagenttable[reagentid] then
						reagenttable[reagentid] = reagenttable[reagentid]..";"..itemid.."x"..reagentnum
					else
						reagenttable[reagentid] = itemid.."x"..reagentnum
					end
					newrecipemats = newrecipemats..reagentid.."x"..reagentnum..";"
				end
				newrecipemats = newrecipemats:sub(1,-2)
				
				if fp_set then
					fp_set = fp_set..","..newrecipemats
				else
					fp_set = "o,"..newrecipemats
				end
			end
			if lp_set then
				lp_set = lp_set..",-"..spellid..":"..itemid
			else
				lp_set = "-"..spellid..":"..itemid
			end
		end
	end)
	for k,v in pairs(reagenttable) do
		rp_set = rp_set..","..k..":"..v
	end
	
	sets["TradeskillResultMats.Forward."..profession] = fp_set
	sets["TradeskillResultMats.Reverse."..profession] = rp_set
	sets["Tradeskill.RecipeLinks."..profession] = lp_set
	
	return newset
end

handlers["^Tradeskill%.Gather"] = function (set, data) -- fixed
	local count = 0
	local gathertype = set:match("%.([^%.]+)$")
	local filter = Tradeskill_Gather_filters[gathertype]
	if filter then
		return basic_listview_handler("http://www.wowhead.com/?items&filter=cr="..filter..";crs=1;crv=0")
	end
end

handlers["^Tradeskill%.Recipe%."] = function (set, data) -- fixed
	local count = 0
	local profession, filter = set:match("^Tradeskill%.Recipe%.([^%.]+)%.(.+)$")
	profession = Tradeskill_Recipe_filters[profession]
	filter = Tradeskill_Recipe_filters[filter]
	if not profession or not filter then return end
	
	local url = "http://www.wowhead.com/?items=9."..profession.."&filter=cr="..filter
	
	return basic_listview_handler(url, function (itemstring)
		local itemid = itemstring:match("{id:(%d+)")
		local skilllvl = itemstring:match("skill:(%d+)")
		return itemid..":"..skilllvl
	end)
end

handlers["^Tradeskill%.Tool"] = function (set, data) -- fixed
	local newset
	local count = 0
	local profession = set:match("^Tradeskill%.Tool%.(.+)$")
	local urls = Tradeskill_Tool_urls[profession]
	if not urls then return end
	
	for _, url in ipairs(urls) do
		local nset = basic_listview_handler(url)
		if nset then
			if newset then
				newset = newset..","..nset
			else
				newset = nset
			end
		end
	end
	return newset
end

handlers["^Tradeskill%.Mat%.ByProfession"] = function (set, data) -- fixed
	local profession = set:match("^Tradeskill%.Mat%.ByProfession%.(.+)$")
	local filter = Tradeskill_Profession_filters[profession]
	if not filter then return end
	local reagentlist = {}
	
	basic_listview_handler("http://www.wowhead.com/?spells="..filter, function (itemstring)
		local reagents = itemstring:match("reagents:%b[]")
		if not reagents then return end
		for r in reagents:gmatch("%[(%d+),%d+%]") do
			reagentlist[tonumber(r)] = true
		end
	end)
	local newset = {}
	for reagent in pairs(reagentlist) do
		newset[#newset + 1] = reagent
	end
	table.sort(newset)
	return table.concat(newset, ",")
end

handlers["^Tradeskill%.Gem%.Cut$"] = function (set, data) -- fixed
	local newset
	local gems = {}	
	basic_listview_handler("http://www.wowhead.com/?items&filter=cr=81;crs=5;crv=0", function (itemstring)
		local itemid = itemstring:match("{id:(%d+)")
		local page = getpage("http://www.wowhead.com/?item="..itemid)
		for list in page:gmatch("new Listview(%b())") do
			if list:match("id: 'created%-by'") then
				local itemlist = list:match("data: %b[]")
				for itemstring in itemlist:gmatch("%b{}") do
					local reagents = itemstring:match("reagents:%b[]")
					for src_id, count in reagents:gmatch("%[(%d+),(%d+)%]") do
						if src_id ~= "27860" then -- Purified Draenic Water 
							local src = gems[src_id] or ""
							gems[src_id] = src..";"..itemid
						end
					end
				end
			end
		end
	end)
	for k, v in pairs(gems) do
		if newset then
			newset = newset..","..k..":"..v:sub(2)
		else
			newset = "o,"..k..":"..v:sub(2)
		end
	end
	return newset
end

handlers["^Tradeskill%.Gem"] = function (set, data) -- fixed
	local color = set:match("%.([^%.]+)$")
	local filter = Tradeskill_Gem_Color_filters[color]
	if not filter then return end
	return basic_listview_handler("http://www.wowhead.com/?items=3."..filter)
end

handlers["^Misc%.Minipet%.Normal"] = function (set, data) -- fixed
	local newset
	local count = 0
	local page = getpage("http://www.wowhead.com/?items=15.-2")
	local itemlist = page:match("new Listview(%b())"):match("data%: (%b[])")
	for itemstring in itemlist:gmatch("%b{}") do
		local itemid = itemstring:match("{id:(%d+)")
		if newset then
			newset = newset..","..itemid
		else
			newset = itemid
		end
		count = count + 1
	end
	return newset
end

handlers["^Misc%.Lockboxes"] = function (set, data) -- fixed
	return basic_listview_handler("http://www.wowhead.com/?items&filter=cr=10;crs=1;crv=0", function (itemstring)
		local itemid = itemstring:match("{id:(%d+)")
		page = getpage("http://www.wowhead.com/?item="..itemid)
		local skill = page:match("Requires Lockpicking %((%d+)%)")
		return itemid..":"..skill
	end)
end

handlers["^Gear%.Socketed"] = function (set, data) -- fixed
	local newset
	local slot = set:match("%.([^%.]+)$")
	for _,url in ipairs(Gear_Socketed_urls[slot]) do
		local nset = basic_listview_handler(url)
		if nset then
			if newset then
				newset = newset..","..nset
			else
				newset = nset
			end
		end
	end
	return newset
end

handlers["Container%.ItemsInType"] = function (set, data) -- fixed
	local newset
	local container = set:match("%.([^%.]+)$")
	local container_id = Containers_ItemsInType_items[container]
	if not container_id then return end
	local page = getpage("http://www.wowhead.com/?item="..container_id)
	for list in page:gmatch("new Listview(%b())") do
		if list:match("id: 'can%-contain'") then
			local itemlist = list:match("data: %b[]")
			for itemstring in itemlist:gmatch("%b{}") do
				local itemid = itemstring:match("{id:(%d+)")
				if newset then
					newset = newset..","..itemid
				else
					newset = itemid
				end
			end
		end
	end
	return newset
end

local function update_all_sets(sets, setcount)
	local setid = 0
	local notmined = {}
	for set, data in pairs(sets) do
		setid = setid + 1
		dprint(1, ("current set: %4d/%4d"):format(setid, setcount), set)
		local newset
		if data:sub(1,2) ~= "m," then
			for pattern, handler in pairs(handlers) do
				if set:match(pattern) then
					local status, result = pcall(handler, set, data)
					if status then
						newset = result
						if result then 
							newset = result
							break
						end
					else
						dprint(1, "ERR", set, pattern, result)
					end
				end
			end
		end
		if newset then
			sets[set] = newset
		else
			table.insert(notmined, set)
		end
	end
	return notmined
end

local function write_output(file, sets)
	local f = assert(io.open(SOURCE, "w"))
	for line in file:gmatch('([^\n]-\n)') do
		local setname, spaces, comment = line:match('\t%[%"([^"]+)%"%]([^=]-)= "[^"]-",([^\n]-)\n')
		if setname and sets[setname] then
			f:write('\t["'..setname..'"]'..spaces..'= "'..sets[setname]..'",'..comment..'\n')
		else
			f:write(line)
		end
	end
	f:close()
end

local function main()
	local starttime = os.time()

	local file, setcount 
	file, sets, setcount = read_data_file()
	local old_enchant = sets["Tradeskill.Crafted.Enchanting"]
	if old_enchant then sets["Tradeskill.Crafted.Enchanting"] = "" end
	local notmined = update_all_sets(sets, setcount)
	if old_enchant then sets["Tradeskill.Crafted.Enchanting"] = old_enchant end
	local elapsed = os.time()- starttime
	local cputime = os.clock()
	print(("Elapsed Time: %dm %ds"):format(elapsed/60, elapsed%60))
	print(("%dm %ds spent servicing %d web requests"):format(httptime/60, httptime%60, httpcount))
	print(("%dm %ds spent in processing data"):format((elapsed-httptime)/60,(elapsed-httptime)%60))
	print(("Approx %dm %.2fs CPU time used"):format(cputime/60, cputime%60))
	local notminedcount = 0
	for k,v in ipairs(notmined) do
		--print("not mined:"..v)
		notminedcount = notminedcount + 1
	end
	print(("%d sets mined, %d sets not mined."):format(setcount-notminedcount, notminedcount))
	write_output(file, sets)
end

main()
